from typing import List, Optional, Any
from pathlib import Path
from urllib.parse import ParseResult
from .uri import path_from_uri
from . import proxy
from .. import missing_data
from silx.utils.proxy import docstring


class FileProxy(proxy.DataProxy, register=False):
    """Example root URI's:
    * "file://path/to/directory"
    * "file://path/to/name.ext"
    * "file://path/to/name.ext?path=/path/in/file"
    """

    EXTENSIONS = NotImplemented
    """Valid file extensions. Expected to be an iterable"""
    ALLOW_PATH_IN_FILE = NotImplemented
    """Does format allow path inside files. Expected to be a boolean"""
    SEP_IN_FILE = "/"
    """Separator used inside files. Expected to be a string"""

    @property
    def path(self) -> Optional[Path]:
        """return 'path' from the following URI representation:
        URI = scheme ":" "//" path ["?" query] ["#" fragment]
        """
        if self.is_fixed_uri:
            return path_from_uri(self.uri.parse())
        parsed_root_uri = self.parsed_root_uri
        if parsed_root_uri is None:
            return None
        identifier = self.identifier
        if identifier is None:
            return None

        # Directory or file name
        root_path = path_from_uri(parsed_root_uri)
        extension = root_path.suffix
        path_is_file = bool(extension)
        if extension not in self.EXTENSIONS:
            extension = self.EXTENSIONS[0]
        path = root_path.with_suffix("")

        # Add path-in-file to path when the file format does not support it
        if self.ALLOW_PATH_IN_FILE:  # for example JSON
            add_identifier_to_path = not path_is_file
        else:  # for example HDF5
            subdirs = self._path_in_file_parts()
            if subdirs:
                path = path.joinpath(*subdirs)
                add_identifier_to_path = False
            else:
                add_identifier_to_path = True

        # Add the identifier to the path if needed
        if add_identifier_to_path:
            filename = identifier
            path /= filename

        return path.with_suffix(extension)

    def _path_in_file_parts(self) -> List[str]:
        parts = [s for s in self._root_uri_path_in_file.split(self.SEP_IN_FILE) if s]
        if self.is_fixed_uri:
            return parts
        identifier = self.identifier
        if identifier is not None:
            parts.append(identifier)
        return parts

    @property
    def _root_uri_path_in_file(self) -> str:
        """
        return '**data root** path query' result
        for data at "file://path/to/name.ext?path=/path/in/file" return "/path/in"
        """
        return self.root_uri_query.get("path", "")

    def path_in_file_parts(self) -> Optional[List[str]]:
        """Return list of 'path' query parts as a LIST WITHOUT the root"""
        if self.ALLOW_PATH_IN_FILE:
            return self._path_in_file_parts()
        else:
            return None

    @property
    def path_in_file(
        self,
    ) -> Optional[str]:
        """
        return '**data** path query' result
        for data at "file://path/to/name.ext?path=/path/in/file" return "path/in/file"
        """
        if self.ALLOW_PATH_IN_FILE:
            return self.SEP_IN_FILE.join(self._path_in_file_parts())
        else:
            return None

    @property
    def path_in_file_parent(
        self,
    ) -> Optional[str]:
        """
        return '**data** path query' result
        for data at "file://path/to/name.ext?path=/path/in/file" return "path/in"
        """
        if self.ALLOW_PATH_IN_FILE:
            parts = self._path_in_file_parts()[:-1]
            return self.SEP_IN_FILE.join(parts)
        else:
            return None

    @property
    def path_in_file_name(self) -> Optional[str]:
        """
        Return '**data** path query' last part
        for data at "file://path/to/name.ext?path=/path/in/file" return "file"
        """
        if self.ALLOW_PATH_IN_FILE:
            return self._path_in_file_parts()[-1]
        else:
            return None

    def _generate_uri(self) -> Optional[proxy.DataUri]:
        path = self.path
        if path is None:
            return
        query = dict()
        path_in_file = self.path_in_file
        if path_in_file:
            query["path"] = path_in_file
        if query:
            query = "&".join([f"{name}={value}" for name, value in query.items()])
        else:
            query = ""

        uri = ParseResult(self.SCHEME, str(path), "", "", query, "")
        return proxy.DataUri(uri, self.uhash)

    @docstring(proxy.DataProxy)
    def exists(self) -> bool:
        path = self.path
        if path is None:
            return False
        return path.exists()

    @docstring(proxy.DataProxy)
    def dump(self, data, **kw):
        path = self.path
        if path is None:
            return False
        self._dump(path, data, **kw)
        return True

    @docstring(proxy.DataProxy)
    def load(self, raise_error=True, **kw):
        path = self.path
        if path is None:
            return missing_data.MISSING_DATA
        try:
            return self._load(path, **kw)
        except FileNotFoundError as e:
            if raise_error:
                raise proxy.UriNotFoundError(path) from e
        except Exception as e:
            if raise_error:
                raise proxy.PersistenceError(path) from e
        return missing_data.MISSING_DATA

    def _dump(self, path: Path, data: Any) -> None:
        raise NotImplementedError

    def _load(self, path: Path) -> Any:
        raise NotImplementedError



from . import clilogutils
from . import cliexecuteutils
from . import cliconvertutils


def add_execute_parameters(parser, shell=False):
    if shell:
        clilogutils.add_log_parameters(parser)
    cliexecuteutils.add_execute_parameters(parser)


def apply_execute_parameters(args, shell=False):
    if shell:
        clilogutils.apply_log_parameters(args)
    cliexecuteutils.apply_execute_parameters(args)


def add_convert_parameters(parser, shell=False):
    if shell:
        clilogutils.add_log_parameters(parser)
    cliconvertutils.add_convert_parameters(parser)


def apply_convert_parameters(args, shell=False):
    if shell:
        clilogutils.apply_log_parameters(args)
    cliconvertutils.apply_convert_parameters(args)



import json
import logging
from typing import Optional

try:
    from papermill import execute_notebook
except ImportError:
    execute_notebook = None

from .task import Task

NOTEBOOK_ARGUMENT = "_notebook"
RESULT_TAG = "results"

logger = logging.getLogger(__name__)


class NotebookExecutorTask(
    Task,
    input_names=[NOTEBOOK_ARGUMENT],
    optional_input_names=["_output_notebook", "_kernel_name", "_execution_timeout"],
    output_names=["output_notebook", "results"],
):
    NOTEBOOK_ARGUMENT = NOTEBOOK_ARGUMENT

    def run(self):
        if execute_notebook is None:
            raise ImportError("requires papermill")
        parameters = self.get_input_values()
        reserved = self.input_names()
        parameters = {k: v for k, v in parameters.items() if k not in reserved}
        kernel_name = self.get_input_value("_kernel_name", None)
        output_notebook = self.get_input_value("_output_notebook", None)
        execution_timeout = self.get_input_value("_execution_timeout", None)
        self.outputs.results = _execute_notebook(
            self.inputs._notebook,
            parameters,
            output_notebook=output_notebook,
            kernel_name=kernel_name,
            execution_timeout=execution_timeout,
        )
        self.outputs.output_notebook = output_notebook


def _execute_notebook(
    input_notebook: str,
    parameters: dict,
    output_notebook: Optional[str] = None,
    kernel_name: Optional[str] = None,
    execution_timeout: Optional[float] = None,
) -> dict:
    level = logger.getEffectiveLevel()
    progress_bar = level and level <= logging.INFO
    nb = execute_notebook(
        input_notebook,
        output_notebook,
        parameters=parameters,
        kernel_name=kernel_name,
        progress_bar=progress_bar,
        execution_timeout=execution_timeout,
    )
    return _extract_results(nb)


def _extract_results(nb) -> dict:
    results = dict()
    for cell in nb.cells:
        metadata = cell["metadata"]
        if "tags" not in metadata or RESULT_TAG not in metadata["tags"]:
            continue
        for output in cell["outputs"]:
            if output["output_type"] != "execute_result":
                continue
            adict = _decode_data(cell, output["data"])
            results.update(adict)
    return results


def _decode_data(cell, data: dict) -> dict:
    jsondata = data.get("application/json")
    if jsondata:
        return jsondata

    txtdata = data.get("text/plain")
    if not txtdata:
        return dict()
    try:
        # Try JSON decoding (e.g. for numbers)
        results = json.loads(txtdata)
    except Exception:
        results = txtdata

    if isinstance(results, dict):
        return results
    key = cell["source"].splitlines()[-1]
    return {key: results}



from typing import Any, Mapping, MutableMapping
from pathlib import Path
import json

from .file import FileProxy
from . import atomic


def modify_dict(target: Mapping, source: MutableMapping):
    for name, value in source.items():
        if isinstance(value, dict):
            new_target = target.setdefault(name, dict())
            modify_dict(new_target, value)
        else:
            target[name] = value


class JsonProxy(FileProxy):
    SCHEME = "json"
    EXTENSIONS = [".json"]
    ALLOW_PATH_IN_FILE = False

    def _dump(self, path: Path, data: Any, **_):
        with atomic.atomic_write(path) as f:
            json.dump(data, f)

    def _load(self, path: Path):
        with open(path, mode="r") as f:
            return json.load(f)



import json
import yaml
import pytest
from ewokscore.graph import load_graph


@pytest.mark.parametrize("with_ext", [True, False])
@pytest.mark.parametrize("representation", ["json", "yaml", "json_module", None])
def test_graph_discovery(representation, with_ext, tmpdir):
    subgraph = {
        "graph": {"id": "subgraph", "input_nodes": [{"id": "in", "node": "subnode1"}]},
        "nodes": [
            {
                "id": "subnode1",
                "task_type": "method",
                "task_identifier": "dummy",
                "default_inputs": [
                    {"name": "name", "value": "subnode1"},
                    {"name": "value", "value": 0},
                ],
            }
        ],
    }

    graph = {
        "graph": {"id": "graph"},
        "nodes": [
            {
                "id": "node1",
                "task_type": "method",
                "task_identifier": "dummy",
                "default_inputs": [
                    {"name": "name", "value": "node1"},
                    {"name": "value", "value": 0},
                ],
            },
            {"id": "node2", "task_type": "graph", "task_identifier": "subgraph"},
        ],
        "links": [
            {
                "source": "node1",
                "target": "node2",
                "sub_target": "in",
                "data_mapping": [
                    {"target_input": "value", "source_output": "return_value"}
                ],
            }
        ],
    }

    ext = ""
    source = "graph"
    root_dir = None
    root_module = None
    if representation == "yaml":
        root_dir = str(tmpdir)
        dump = yaml.dump
        if with_ext:
            ext = ".yml"
    elif representation == "json":
        root_dir = str(tmpdir)
        dump = json.dump
        if with_ext:
            ext = ".json"
    elif representation == "json_module":
        root_module = "ewokscore.tests.examples.loadtest"
        if with_ext:
            source = root_module + "." + source
        else:
            representation = None
        dump = None
    else:
        root_dir = str(tmpdir)
        dump = json.dump
        if with_ext:
            ext = ".json"

    if dump is not None:
        destination = str(tmpdir / "subgraph" + ext)
        with open(destination, mode="w") as f:
            dump(subgraph, f)
        destination = str(tmpdir / "graph" + ext)
        with open(destination, mode="w") as f:
            dump(graph, f)

    ewoksgraph = load_graph(
        source,
        representation=representation,
        root_dir=root_dir,
        root_module=root_module,
    )

    assert set(ewoksgraph.graph.nodes) == {"node1", ("node2", "subnode1")}



import numpy
from typing import Any, Dict, Union, Optional
from urllib.parse import ParseResult

from .uri import parse_query, parse_uri, join_uri, uri_as_string
from ..registration import Registered
from ..hashing import UniversalHash
from ..hashing import HasUhash


class PersistenceError(RuntimeError):
    pass


class UriNotFoundError(PersistenceError):
    pass


class DataUri(HasUhash):
    def __init__(self, uri: str, uhash: Union[UniversalHash, str]):
        if isinstance(uri, numpy.ndarray):
            uri = uri.item()
        if isinstance(uhash, numpy.ndarray):
            uhash = uhash.item()
        self.__uri = uri_as_string(uri)
        if isinstance(uhash, str):
            uhash = UniversalHash(uhash)
        self.__uhash = uhash

    def __repr__(self):
        return f"{type(self).__name__}({self.__uri})"

    def __str__(self):
        return self.__uri

    def parse(self) -> ParseResult:
        return parse_uri(self.__uri)

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == str(other)
        elif isinstance(other, DataUri):
            return str(self) == str(other) and self.uhash == other.uhash
        else:
            return False

    def __copy__(self):
        return type(self)(self.__uri, self.__uhash)

    @property
    def uhash(self) -> UniversalHash:
        return self.__uhash

    def serialize(self) -> Dict[str, str]:
        return {"uri": self.__uri, "uhash": str(self.uhash)}

    @classmethod
    def deserialize(cls, data: Dict[str, str]):
        if not isinstance(data, dict):
            return None
        try:
            uri = data["uri"]
            uhash = data["uhash"]
        except KeyError:
            return None
        return cls(uri, uhash)


class DataProxy(Registered, HasUhash, register=False):
    SCHEME = NotImplemented
    """name of the DataProxy scheme like json or nexus"""

    def __init__(
        self,
        uri: Optional[DataUri] = None,
        root_uri: Optional[str] = None,
        relative_uri: Optional[str] = None,
        uhash_source: Optional[Union[UniversalHash, HasUhash]] = None,
    ):
        """Either the URI is provided or the root + relative URI with a uhash source (the URI can be derived from those)"""
        self.__parsed_root_uri = None
        self.__fixed_uri = uri
        self.__uhash_source = uri
        if uri is None:
            self.__uhash_source = uhash_source
        else:
            root_uri = str(uri)
            relative_uri = None
        if root_uri:
            parsed_root_uri = parse_uri(root_uri)
            if relative_uri:
                parsed_root_uri = join_uri(parsed_root_uri, relative_uri)
            self.__parsed_root_uri = parsed_root_uri

    def __repr__(self):
        uri = self.uri
        if uri is None:
            return super().__repr__()
        else:
            return f"{super().__repr__()}(uri='{uri}')"

    @classmethod
    def instantiate(
        cls,
        scheme: Optional[str] = None,
        uri: Optional[DataUri] = None,
        uhash_source: Optional[Union[UniversalHash, HasUhash]] = None,
        root_uri: Optional[str] = None,
        relative_uri: Optional[str] = None,
    ):
        if uri is not None:
            scheme = uri.parse().scheme
        elif isinstance(root_uri, DataProxy):
            scheme = root_uri.SCHEME
        elif isinstance(root_uri, DataUri):
            scheme = root_uri.parse().scheme
        for subclass in cls.get_subclasses():
            if subclass.SCHEME == scheme:
                return subclass(
                    uri=uri,
                    uhash_source=uhash_source,
                    root_uri=root_uri,
                    relative_uri=relative_uri,
                )
        raise ValueError(f"Data proxy scheme '{scheme}' is not supported")

    def serialize(self) -> Dict[str, str]:
        return self.uri.serialize()

    @classmethod
    def deserialize(self, data: Dict[str, str]):
        uri = DataUri.deserialize(data)
        if uri is None:
            return None
        return self.instantiate(uri=uri)

    @property
    def uhash(self) -> Optional[UniversalHash]:
        if self.is_fixed_uri:
            return self.__fixed_uri.uhash
        elif isinstance(self.__uhash_source, HasUhash):
            return self.__uhash_source.uhash
        elif isinstance(self.__uhash_source, UniversalHash):
            return self.__uhash_source
        else:
            return None

    @property
    def identifier(self) -> Optional[str]:
        """Return identifier DataProxy to be used as a string"""
        uhash = self.uhash
        if uhash is None:
            return None
        return str(uhash)

    @property
    def parsed_root_uri(self) -> Optional[ParseResult]:
        return self.__parsed_root_uri

    @property
    def root_uri_query(self) -> dict:
        parsed_root_uri = self.parsed_root_uri
        if parsed_root_uri:
            return parse_query(parsed_root_uri)
        return dict()

    @property
    def is_fixed_uri(self) -> bool:
        return self.__fixed_uri is not None

    @property
    def uri(self) -> Optional[DataUri]:
        """
        Return an Unified Resource Identifier. Defined as:
        URI = scheme ":" "//" path ["?" query] ["#" fragment]

        see https://en.wikipedia.org/wiki/Uniform_Resource_Identifier

        .. warning:: query can be ?path= which is different from path
        """
        if self.is_fixed_uri:
            return self.__fixed_uri
        return self._generate_uri()

    def _generate_uri(self) -> Optional[DataUri]:
        """Generate a URI based on the root URI and the universal hash"""
        raise NotImplementedError

    def exists(self) -> bool:
        """return True if the data exists"""
        raise NotImplementedError

    def load(self, raise_error: bool = True) -> Any:
        """Load data from the uri"""
        raise NotImplementedError

    def dump(self, data: Any) -> bool:
        """Dump data to the uri"""
        raise NotImplementedError



from typing import Optional, Union

NodeIdType = Union[str, int, tuple]

SEPARATOR = ":"


def flatten_node_id(node_id: NodeIdType) -> tuple:
    if isinstance(node_id, str):
        return (node_id,)
    elif isinstance(node_id, int):
        return str(node_id)
    elif len(node_id) == 1:
        return (node_id[0],)
    else:
        return (node_id[0],) + flatten_node_id(node_id[1])


def node_id_as_string(node_id: NodeIdType, sep: Optional[str] = None) -> str:
    if sep is None:
        sep = SEPARATOR
    return sep.join(flatten_node_id(node_id))


def node_id_from_json(node_id: Union[list, NodeIdType]) -> NodeIdType:
    if isinstance(node_id, list):
        return tuple(map(node_id_from_json, node_id))
    return node_id


def get_node_id(
    node_id: Optional[NodeIdType], node_attrs: Optional[dict]
) -> Optional[NodeIdType]:
    if node_id is None and node_attrs:
        return node_attrs.get("id")
    return node_id


def get_node_label(
    node_id: Optional[NodeIdType], node_attrs: Optional[dict], sep: Optional[str] = None
) -> Optional[str]:
    if node_attrs:
        node_label = node_attrs.get("label")
        if node_label:
            return node_label
    node_id = get_node_id(node_id, node_attrs)
    if node_id:
        return node_id_as_string(node_id, sep=sep)
    return None


def get_varinfo(node_attrs: Optional[dict], varinfo: Optional[dict] = None) -> dict:
    """Makes a copy"""
    if varinfo is None:
        varinfo = dict()
    else:
        varinfo = dict(varinfo)
    if node_attrs:
        varinfo.update(node_attrs.get("varinfo", dict()))
    return varinfo


def get_task_identifier(node_attrs: Optional[dict], default: Optional[str] = None):
    if node_attrs:
        return node_attrs.get("task_identifier", default)
    return default



from typing import Optional
from .task import Task
from .progress import BaseProgress


class TaskWithProgress(Task, register=False):
    """
    Task within a progress to display task advancement
    """

    def __init__(self, *args, **kw):
        self._task_progress: Optional[BaseProgress] = kw.pop("progress", None)
        super().__init__(*args, **kw)

    @property
    def progress(self) -> Optional[int]:
        """Task advancement. If a task progress is not provided then return
        None"""
        if self._task_progress is not None:
            return self._task_progress.progress
        else:
            return None

    @progress.setter
    def progress(self, progress: int):
        if self._task_progress:
            current = self._task_progress.progress
            self._task_progress.progress = progress
            new = self._task_progress.progress
            if current != new:
                self._send_event(event="progress", progress=new)
